#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include "libmadden.h"
#include "libcrc32.h"

static unsigned long resolve_number(tdb_field *f, unsigned char *data);

static tdb_field *resolve_field(tdb_t *tdb, const char *field)
{
  unsigned long fno;
  long floc = -1;

  for (fno = 0; fno < tdb->current.header.field_count; fno++) {
    if (strncmp(tdb->current.fields[fno].name, field, 4) != 0)
      continue;

    floc = fno;
    break;
  }

  if (floc < 0) {
    fprintf(stderr, "tdb::resolve_field: field '%s' not found.\n", field);
    return NULL;
  }
  return &tdb->current.fields[floc];
}


static unsigned long rewrite_number(tdb_field *f, unsigned char *data, unsigned char *writ)
{
  int ibit, obit;
  char imask, omask;
  /* bit-by-bit */
  fprintf(stderr, "  start\t%lu\n", resolve_number(f, data));
  for (	ibit = 0, obit = f->offset, imask = (1 << 0), omask = (1 << (f->offset % 8));
	obit < f->offset + f->bits;
	obit++, ibit++, imask <<= 1, omask <<= 1) {
    if (imask == 0) imask = 1;
    if (omask == 0) omask = 1;
//    fprintf(stderr, "ib=%d ob=%d im=%#04hhx om=%#04hhx ik=%u ok=%u ", ibit, obit, imask, omask, ibit / 8, obit / 8);
//    fprintf(stderr, "%#04x(%u->%u) -> ", data[obit / 8], data[obit / 8] & omask ? 1 : 0, writ[ibit / 8] & imask ? 1 : 0);
    if (writ[ibit / 8] & imask)
      data[obit / 8] |= omask;
    else
      data[obit / 8] &= ~omask;
//    fprintf(stderr, "%#04x\n", data[obit / 8]);
  }
fprintf(stderr, "  end\t%lu\n", resolve_number(f, data));
}


static unsigned long resolve_number(tdb_field *f, unsigned char *data)
{
  unsigned long out_shift = 0, left = f->bits, data_bib, bucket_data,
	mask, value, bitdata = 0,
	in_shift = f->offset % 32, bucket = f->offset / 32;

  while (out_shift < f->bits) {
    // 4 cases (LSB)
    // .....111 - Start of read.
    // 11111111 - Continuing read.
    // 111..... - End of read.
    // ...11... - Single read.
    if (in_shift != 0) {
      if (left + in_shift > 32) {
        // Case 1
        data_bib = 32 - in_shift;
      } else {
        // Case 4
        data_bib = f->bits;
      }
    } else {
      // Case 2 & 3
      data_bib = left > 32 ? 32 : left;
    }

    bucket_data =
		(data[bucket * sizeof(long) + 0] <<  0) +
		(data[bucket * sizeof(long) + 1] <<  8) +
		(data[bucket * sizeof(long) + 2] << 16) +
		(data[bucket * sizeof(long) + 3] << 24);

    mask = (1 << data_bib) - 1;
    value = (bucket_data >> in_shift) & mask;
    bitdata |= value << out_shift;

    left -= data_bib;
    out_shift += data_bib;
    if (in_shift) {
      in_shift = 0;
    }
    bucket++;
  }

  return bitdata;
}


static unsigned long resolve_integer(tdb_field *f, unsigned char *data)
{
  return resolve_number(f, data);
}


tdb_t *tdb_open(const char *filename, const char *filemode)
{
  unsigned long db_id, junk, file_size;
  struct stat st;
  tdb_t *file = calloc(1, sizeof(tdb_t));

  if (!file)
    /* out of memory */
    return NULL;
  else if (!(file->fd = fopen(filename, filemode)))
    fprintf(stderr, "%s: cannot open with mode '%s': %s\n", filename, filemode, strerror(errno));
  else if (fread(&db_id, sizeof(db_id), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading database identifier: %s\n", filename, strerror(errno));
  else if (fread(&junk, sizeof(junk), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading field #2: %s\n", filename, strerror(errno));
  else if (fread(&file_size, sizeof(file_size), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading file size: %s\n", filename, strerror(errno));
  else if (fstat(fileno(file->fd), &st) < 0)
    fprintf(stderr, "%s: error getting file system information: %s\n", filename, strerror(errno));
  else if (st.st_size != file_size)
    fprintf(stderr, "%s: internal file size (%lu) doesn't match file size(%lu)\n", filename, file_size, st.st_size);
  else if (fread(&junk, sizeof(junk), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading field #4: %s\n", filename, strerror(errno));
  else if (fread(&file->table_count, sizeof(file->table_count), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading table count: %s\n", filename, strerror(errno));
  else if (fread(&junk, sizeof(junk), 1, file->fd) != 1)
    fprintf(stderr, "%s: error reading field #6: %s\n", filename, strerror(errno));
  else if ((file->tables = calloc(file->table_count, sizeof(*file->tables))) == NULL)
    goto error_exit;
  else {
    size_t tp;

    for (tp = 0; tp < file->table_count; tp++) {
      if (fread(&file->tables[tp], sizeof(*file->tables), 1, file->fd) == 1) {
//        fprintf(stderr, "tdb_open: %.4s @ %lu\n", file->tables[tp].name, file->tables[tp].offset);
        continue;
      }

      fprintf(stderr, "%s: error reading table #%zu: %s\n", filename, tp, strerror(errno));
      goto error_exit;
    }

    if (fread(&junk, sizeof(junk), 1, file->fd) != 1)
      fprintf(stderr, "%s: error reading final crc: %s\n", filename, strerror(errno));

    file->table_data_start = ftell(file->fd);

    return file;
  }

error_exit:
  if (file->tables)
    free(file->tables);
  if (file->fd)
    fclose(file->fd);
  free(file);
  return NULL;
}

  
int tdb_close(tdb_t *tdb)
{
  if (!tdb)
    return FALSE;
  if (tdb->tables)
    free(tdb->tables);
  if (tdb->current.fields)
    free(tdb->current.fields);
  if (tdb->fd)
    fclose(tdb->fd);
  free(tdb);
  return TRUE;
}


int tdb_write(tdb_t *tdb, const char *field_name, void *writ)
{
  tdb_field *f;

  if (!(f = resolve_field(tdb, field_name)))
    return -1;

  if (fseek(tdb->fd, tdb->current.data_start, SEEK_SET) < 0)
    fprintf(stderr, "tdb_search: error resetting record start: %s\n", strerror(errno));

  else if (fseek(tdb->fd, tdb->current.header.record_length_bytes * tdb->current.cursor_record, SEEK_CUR) < 0)
    fprintf(stderr, "tdb_search: error resetting position: %s\n", strerror(errno));

  else if (f->type == TDB_FIELD_STRING || f->type == TDB_FIELD_BINARY) {
    if (f->offset % 8 != 0) {
      fprintf(stderr, "PANIC: unhandled condition: string/binary field not byte-aligned: %.4s/%.4s\n", tdb->tables[tdb->current.table].name, f->name);
      exit(EXIT_FAILURE);
    }
    fseek(tdb->fd, f->offset / 8, SEEK_CUR);
    fwrite(writ, f->bits / 8, 1, tdb->fd);
    return 0;

  } else if (f->type == TDB_FIELD_SINT || f->type == TDB_FIELD_UINT) {
    char *data;

    if (!(data = calloc(1, tdb->current.header.record_length_bytes)))
      return -1;
    else if (fread(data, tdb->current.header.record_length_bytes, 1, tdb->fd) != 1)
      fprintf(stderr, "tdb_search: error reading record: %s\n", strerror(errno));
    else if (fseek(tdb->fd, -tdb->current.header.record_length_bytes, SEEK_CUR) < 0)
      fprintf(stderr, "tdb_search: error resetting position for write: %s\n", strerror(errno));
    else {
      rewrite_number(f, data, (unsigned char *)writ);
      if (fwrite(data, tdb->current.header.record_length_bytes, 1, tdb->fd) != 1)
        fprintf(stderr, "tdb_search: error writing record: %s\n", strerror(errno));
    }
    if (data)
      free(data);
    return 0;

  } else if (f->type == TDB_FIELD_FLOAT) {
    fprintf(stderr, "PANIC: unhandled condition: floating point number: %.4s/%.4s\n", tdb->tables[tdb->current.table].name, f->name);
    exit(EXIT_FAILURE);
  }

  return -1;
}

int tdb_open_table(tdb_t *tdb, const char *table_name)
{
  size_t tp, table_loc = -1;

  if (!tdb || !table_name)
    return FALSE;

  for (tp = 0; tp < tdb->table_count; tp++) {
    if (strncmp(tdb->tables[tp].name, table_name, 4) != 0)
      continue;

    table_loc = tp;
    break;
  }

  if (table_loc < 0)
    return FALSE;

  tdb->current.table = tp;
//  fprintf(stderr, "tdb_open_table: found %.4s/%d(%lu) + %zu + %u for %s\n", tdb->tables[tp].name, tp, tdb->tables[tp].offset, tdb->table_data_start, sizeof(tdb->current.header), table_name);

  if (fseek(tdb->fd, tdb->table_data_start, SEEK_SET) < 0)
    fprintf(stderr, "table %s: error seeking to file start: %s\n", table_name, strerror(errno));
  else if (fseek(tdb->fd, tdb->tables[tp].offset, SEEK_CUR) < 0)
    fprintf(stderr, "table %s: error seeking to table start: %s\n", table_name, strerror(errno));
  else if (fread(&tdb->current.header, sizeof(tdb->current.header), 1, tdb->fd) != 1)
    fprintf(stderr, "table %s: error reading header: %s\n", table_name, strerror(errno));
  else if (!(tdb->current.fields = calloc(tdb->current.header.field_count, sizeof(*tdb->current.fields))))
    return FALSE;
  else {
    tdb->current.field_start = ftell(tdb->fd);

    for (tp = 0; tp < tdb->current.header.field_count; tp++) {
      if (fread(&tdb->current.fields[tp], sizeof(*tdb->current.fields), 1, tdb->fd) == 1) {
//        fprintf(stderr, "fields[%zu]: %lu %lu %.4s %lu\n", tp, tdb->current.fields[tp].type, tdb->current.fields[tp].offset, tdb->current.fields[tp].name, tdb->current.fields[tp].bits);
        continue;
      }

      fprintf(stderr, "table %s: error reading field #%u: %s\n", table_name, tp, strerror(errno));
      free(tdb->current.fields);
      return FALSE;
    }

    tdb->current.data_start = ftell(tdb->fd);
    tdb->current.cursor_record = 0;

    return TRUE;
  }

  return FALSE;
}

int tdb_seek(tdb_t *tdb, long offset, int whence)
{
  if (!tdb || !tdb->current.data_start)
    return -1;
  else if (tdb->current.cursor_record + offset < 0)
    return -1;
  else if (tdb->current.cursor_record + offset > tdb->current.header.records_used)
    return -1;
  else
/* FIXME XXX: update cursor_record */
    return fseek(tdb->fd, offset * tdb->current.header.record_length_bytes, whence);
}


int tdb_get_fields(tdb_t *tdb, char *fields[])
{
  /* for now, just use tdb->current.fields */
  return 0;
}


long tdb_search(tdb_t *tdb, const char *field, void *value)
{
  size_t rno = 0;
  unsigned char *data = NULL;
  tdb_field *f;

  if (!tdb || !field || !value) {
    fprintf(stderr, "tdb_search: missing input: %p %p %p\n", tdb, field, value);
    return -1;
  }

  if (!(f = resolve_field(tdb, field)))
    return -1;

//fprintf(stderr, "ccr=%zu rlB=%lu rlb=%lu ru=%d ra=%d fos=%lu fln=%lu\n", tdb->current.cursor_record, tdb->current.header.record_length_bytes, tdb->current.header.record_length_bits, tdb->current.header.records_used, tdb->current.header.records_allocated, f->offset, f->bits);
  if (fseek(tdb->fd, tdb->current.data_start, SEEK_SET) < 0)
    fprintf(stderr, "tdb_search: error resetting record start: %s\n", strerror(errno));
  else if (fseek(tdb->fd, tdb->current.header.record_length_bytes * tdb->current.cursor_record, SEEK_CUR) < 0)
    fprintf(stderr, "tdb_search: error resetting position: %s\n", strerror(errno));
  else if (!(data = calloc(1, tdb->current.header.record_length_bytes)))
    return -1;
  else {
    while (tdb->current.cursor_record <= tdb->current.header.records_used) {
      if (fread(data, tdb->current.header.record_length_bytes, 1, tdb->fd) != 1) {
        fprintf(stderr, "tdb_search: error reading record: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
      if (f->type == TDB_FIELD_STRING) {
        if (f->offset % 8 != 0) {
          fprintf(stderr, "PANIC: unhandled condition: string field not byte-aligned: %.4s/%.4s\n", tdb->tables[tdb->current.table].name, f->name);
          exit(EXIT_FAILURE);
        }
//fprintf(stderr, "compare: '%s' & '%.16s'\n", (const char *)value, data + f->offset / 8);
        if (!strncmp((const char *)value, data + f->offset / 8, f->bits / 8))
          goto ret;
      } else if (f->type == TDB_FIELD_BINARY) {
        if (f->offset % 8 != 0) {
          fprintf(stderr, "PANIC: unhandled condition: binary field not byte-aligned: %.4s/%.4s\n", tdb->tables[tdb->current.table].name, f->name);
          exit(EXIT_FAILURE);
        }
        if (!memcmp(value, data + f->offset / 8, f->bits / 8))
          goto ret;
      } else if (f->type == TDB_FIELD_SINT || f->type == TDB_FIELD_UINT) {
        unsigned long valint = *(unsigned long *)value;
        unsigned long dbint = resolve_integer(f, data);
//fprintf(stderr, "%lu ?= %lu\n", valint, dbint);
        if (dbint == valint)
          goto ret;
      } else if (f->type == TDB_FIELD_FLOAT) {
        fprintf(stderr, "PANIC: unhandled condition: floating point number: %.4s/%.4s\n", tdb->tables[tdb->current.table].name, f->name);
        exit(EXIT_FAILURE);
      }

      tdb->current.cursor_record++;
      rno++;
    }
  }
  rno = -1;
ret:
  if (data)
    free(data);
  return rno;
}


int tdb_close_table(tdb_t *tdb)
{
  size_t len;
  unsigned int empty, crc = 0;

  if (!tdb)
    return -1;

  crc32init_be();

  empty = (tdb->current.header.UNKNOWN_1 >> 24) & 0x10;
  len = tdb->current.header.field_count * sizeof(long) * 4
	+ (!empty ? tdb->current.header.record_length_bytes * tdb->current.header.records_allocated
		: 0)
	+ tdb->current.header.index_count * sizeof(long) * 4;
  len /= 4;
//fprintf(stderr, "crclen=%zu empty=%u\n", len, empty);

  fseek(tdb->fd, tdb->current.field_start, SEEK_SET);
//fprintf(stderr, "here=%lu rlb=%lu rac=%lu ic=%lu\n", ftell(tdb->fd), tdb->current.header.record_length_bytes, tdb->current.header.records_allocated, tdb->current.header.index_count);
  while (len--) {
    unsigned long word;
    fread(&word, sizeof(word), 1, tdb->fd);
    crc = crc32_be(crc, (unsigned char *)&word, sizeof(word));
  }
  crc ^= 0xFFFFFFFF;

//  fprintf(stderr, "crc @ %lu\n", ftell(tdb->fd));
  if (1) {
    unsigned int oldcrc;
    fread(&oldcrc, sizeof(oldcrc), 1, tdb->fd);
    fseek(tdb->fd, -(signed int)sizeof(oldcrc), SEEK_CUR);
    printf("old crc: %#010x\nnew crc: %#010x\n", oldcrc, crc);
  }
  fwrite(&crc, sizeof(crc), 1, tdb->fd);

  return 0;
}
